@@ -4,6 +4,8 @@
 
 #include "util/testharness.h"
 
+#include <string>
+#include <stdlib.h>
 #include <sys/stat.h>
 #include <sys/types.h>
 
@@ -32,10 +34,20 @@ bool RegisterTest(const char* base, const char* name, void (*func)()) {
 }
 
 int RunAllTests() {
+  const char* matcher = getenv("LEVELDB_TESTS");
+
   int num = 0;
   if (tests != NULL) {
     for (int i = 0; i < tests->size(); i++) {
       const Test& t = (*tests)[i];
+      if (matcher != NULL) {
+        std::string name = t.base;
+        name.push_back('.');
+        name.append(t.name);
+        if (strstr(name.c_str(), matcher) == NULL) {
+          continue;
+        }
+      }
       fprintf(stderr, "==== Test %s.%s\n", t.base, t.name);
       (*t.func)();
       ++num;
